import {
  CMType,
  ComponentRequest,
  HMType,
  HostMessage,
  ValidationError,
} from '../types';
import {
  MessageCallback,
  MessageEventManager,
} from '../utils/MessageEventManager';

/**
 * An embedded Notabene component
 * @public
 */

export default class EmbeddedComponent<V, R> {
  private _url: string;
  private _value: V;
  private _errors: ValidationError[] = [];

  private iframe?: HTMLIFrameElement;
  private eventManager: MessageEventManager;
  private modal?: HTMLDialogElement;

  /**
   * Creates an instance of EmbeddedComponent.
   * @param url - The URL of the embedded component
   * @param value - The initial transaction value
   */

  constructor(url: string, value: V) {
    this._url = url;
    this._value = value;
    this.eventManager = new MessageEventManager();
    this.on(CMType.INVALID, (message) => {
      if (message.type === CMType.INVALID) {
        this._errors = message.errors;
        this._value = message.value as V;
      }
    });
    this.on(CMType.RESIZE, (message) => {
      if (message.type === CMType.RESIZE) {
        if (this.iframe) this.iframe.style.height = `${message.height}px`;
      }
    });
  }
  /**
   * Gets the URL of the embedded component
   * @returns The URL of the embedded component
   */

  public get url(): string {
    return this._url;
  }
  /**
   * Gets the current transaction value
   * @returns The current transaction value
   */

  public get value(): V {
    return this._value;
  }

  public get errors(): ValidationError[] {
    return this._errors;
  }

  /**
   * Opens the component URL in the current window
   */

  public open() {
    document.location.href = this.url;
  }

  /**
   * Mounts the component to a parent element
   * @param parentId - The ID of the parent element
   * @throws Will throw an error if the parent element is not found
   */

  mount(parentId: string) {
    const parent = document.querySelector(parentId);
    if (!parent) throw new Error(`parentID ${parentId} not found`);
    this.embed(parent);
  }
  /**
   * Embeds the component into a parent element
   * @param parent - The parent element to embed the component into
   */

  embed(parent: Element, modal: boolean = false) {
    // console.log(this.url);
    this.iframe = document.createElement('iframe');
    this.iframe.src = this.url + (!modal ? '&embedded=true' : '');
    this.iframe.allow = 'web-share';
    this.iframe.style.width = '100%';
    this.iframe.style.height = '100px';
    this.iframe.style.border = 'none';
    this.iframe.style.overflow = 'hidden';
    // this.iframe.setAttribute('allowtransparency', 'true');
    // this.iframe.style.backgroundColor = 'transparent';
    parent.appendChild(this.iframe);

    window.addEventListener('message', (event: MessageEvent) => {
      if (event.source !== this.iframe?.contentWindow) {
        return;
      }
      // console.log('received message from iframe', event.data);
      this.eventManager?.setPort(event.ports[0]);
    });

    this.iframe?.contentWindow?.focus();
  }

  /**
   * Sends a message to the embedded component
   * @param message - The message to send
   */
  send(message: HostMessage): void {
    this.eventManager.send(message);
  }

  /**
   * Adds an event listener for a specific message type
   * @param messageType - The type of message to listen for
   * @param callback - The callback function to execute when the message is received
   */

  on(messageType: string, callback: MessageCallback): void {
    this.eventManager.on(messageType, callback);
  }

  /**
   * Removes an event listener for a specific message type
   * @param messageType - The type of message to stop listening for
   * @param callback - The callback function to remove
   */

  off(messageType: string, callback: MessageCallback): void {
    this.eventManager.off(messageType, callback);
  }

  /**
   * Updates the transaction value and sends an update message to the component
   * @param value - The new transaction value
   */
  update(value: V) {
    this._value = value;
    this.send({ type: HMType.UPDATE, value: value as ComponentRequest });
  }

  /**
   * Waits for the component to complete and returns the transaction response
   * @returns A promise that resolves with the transaction response
   */

  completion(): Promise<R> {
    return new Promise((resolve, reject) => {
      this.on(CMType.COMPLETE, (message) => {
        resolve((message as any).response);
      });
      this.on('error', (message) => {
        reject(new Error((message as any).message));
      });
    });
  }

  /**
   * Opens the component in a modal dialog
   */

  openModal(): Promise<R> {
    if (this.modal) {
      this.closeModal();
    }
    this.modal = document.createElement('dialog');
    this.modal.style.border = 'none';
    this.modal.style.backgroundColor = 'transparent';
    this.modal.style.maxWidth = '100vw';
    this.modal.style.maxHeight = '100vh';
    this.modal.style.width = '600px';
    this.modal.style.height = '800px';

    document.body.appendChild(this.modal);
    this.embed(this.modal, true);
    this.on(CMType.CANCEL, () => {
      this.closeModal();
    });
    this.on(CMType.CLOSE, () => {
      this.closeModal();
    });
    this.on(CMType.COMPLETE, () => {
      this.closeModal();
    });

    this.modal.showModal();
    this.modal.addEventListener('click', () => {
      // console.log('click');
      this.closeModal();
    });
    return this.completion();
  }
  /**
   * Closes the modal dialog
   *
   */

  closeModal(): void {
    if (this.modal) {
      // console.log('closeModal');
      this.modal?.close();
      this.modal.remove();
      this.modal = undefined;
    }
  }
}
